// qtractorObserver.cpp
//
/****************************************************************************
   Copyright (C) 2005-2009, rncbc aka Rui Nuno Capela. All rights reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*****************************************************************************/

#include "qtractorObserver.h"


//---------------------------------------------------------------------------
// qtractorSubject - Scalar parameter value model.

// Constructor.
qtractorSubject::qtractorSubject ( const QString& sName, float fValue )
	: m_sName(sName), m_fValue(fValue), m_bBusy(false)
{
}

// Destructor.
qtractorSubject::~qtractorSubject (void)
{
	m_bBusy = true;

	QListIterator<qtractorObserver *> iter(m_observers);
	while (iter.hasNext())
		iter.next()->setSubject(NULL);

	m_observers.clear();
}


// Subjecteter name accessors.
void qtractorSubject::setName ( const QString& sName )
{
	m_sName = sName;
}

const QString& qtractorSubject::name (void) const
{
	return m_sName;
}


// Direct value accessors.
void qtractorSubject::setValue ( float fValue, qtractorObserver *pSender )
{
	float fOldValue = m_fValue;

	m_fValue = fValue;

	if (fValue != fOldValue)
		notify(pSender);
}

float qtractorSubject::value (void) const
{
	return m_fValue;
}


// Busy flag predicate.
bool qtractorSubject::isBusy() const
{
	return m_bBusy;
}


// Observer/view updater.
void qtractorSubject::notify ( qtractorObserver *pSender )
{
	m_bBusy = true;

	QListIterator<qtractorObserver *> iter(m_observers);
	while (iter.hasNext()) {
		qtractorObserver *pObserver = iter.next();
		if (pSender && pSender == pObserver)
			continue;
		pObserver->update();
	}

	m_bBusy = false;
}


// Observer list accessors.
void qtractorSubject::attach ( qtractorObserver *pObserver )
{
	m_observers.append(pObserver);
}

void qtractorSubject::detach ( qtractorObserver *pObserver )
{
	int iIndex = m_observers.indexOf(pObserver);
	if (iIndex >= 0)
		m_observers.removeAt(iIndex);
}


const QList<qtractorObserver *>& qtractorSubject::observers (void) const
{
	return m_observers;
}


//---------------------------------------------------------------------------
// qtractorObserver - Scalar parameter value control/view.

// Constructor.
qtractorObserver::qtractorObserver ( qtractorSubject *pSubject )
	: m_pSubject(pSubject)
{
	if (m_pSubject) m_pSubject->attach(this);
}

// Virtual destructor.
qtractorObserver::~qtractorObserver (void)
{
	if (m_pSubject)	m_pSubject->detach(this);
}


// Subject value accessor.
void qtractorObserver::setSubject ( qtractorSubject *pSubject )
{
	if (m_pSubject && pSubject)
		m_pSubject->detach(this);
	
	m_pSubject = pSubject;

	if (m_pSubject)
		m_pSubject->attach(this);
}

qtractorSubject *qtractorObserver::subject (void) const
{
	return m_pSubject;
}


// Indirect value accessors.
void qtractorObserver::setValue ( float fValue )
{
	if (m_pSubject) m_pSubject->setValue(fValue, this);
}

float qtractorObserver::value (void) const
{
	return (m_pSubject ? m_pSubject->value() : 0.0f);
}


// Busy flag predicate.
bool qtractorObserver::isBusy (void) const
{
	return (m_pSubject ? m_pSubject->isBusy() : true);
}


// end of qtractorObserver.cpp
